/* handlers.H
 * John Viega
 *
 * Jan 28-29 2000
 */

#ifndef __HANDLERS_H__
#define __HANDLERS_H__

#include "lex.H"
#include "vulndb.H"
#include "toctou.H"

void InitHandlers();
void DoPostProcessing();
void DefaultHandler(VulnInfo *, TokenContainer *, int, char *);      // 0
void StrcpyHandler(VulnInfo *, TokenContainer *, int, char *);       // 1
void SprintfHandler(VulnInfo *, TokenContainer *, int, char *);      // 2
void SnprintfHandler(VulnInfo *, TokenContainer *, int, char *);     // 3
void ScanfHandler(VulnInfo *, TokenContainer *, int, char *);        // 4
void SscanfHandler(VulnInfo *, TokenContainer *, int, char *);       // 5
void TOCTOU_A_Handler(VulnInfo *, TokenContainer *, int, char *);    // 6
void TOCTOU_B_Handler(VulnInfo *, TokenContainer *, int, char *);    // 7
void TOCTOU_C_Handler(VulnInfo *, TokenContainer *, int, char *);    // 8
void FprintfHandler(VulnInfo *, TokenContainer *, int, char *);      // 9
void PrintfHandler(VulnInfo *, TokenContainer *, int, char *);       // 10
void SyslogHandler(VulnInfo *, TokenContainer *, int, char *);       // 11
#endif
